// **********************************************************************************************************
// *																										*
// * This Software is published under the MIT License.														*
// * For more information and the license text, please refer to the file 'License' located in the zip file. *
// *																										*
// * Copyright (c) 2001-2011 Steffen Ernsting, Philipp Ciechanowicz, Michael Poldner and Herbert Kuchen     *                 
// *        		{s.ernsting|ciechanowicz|poldner|kuchen}@uni-muenster.de              					*
// *																										*	
// **********************************************************************************************************

#include <stdlib.h>

#include "TestBitonic.h"
#include "DistributedArray.h"
#include "Muesli.h"
#include "OAL.h"
#include "Utility.h"

int msl::test::bitonic::random(int i) {
	return rand() % 100;
}

bool msl::test::bitonic::odd(int i) {
	return (i & 1) != 0;
}

int msl::test::bitonic::bitsum(int i) {
	int sum = 0;
	
	while(i > 0) {
		sum += (i & 1);
		i >>= 1;
	}
	
	return sum;
}

bool msl::test::bitonic::initDirection(int i) {
	return odd(bitsum(i >> 1));
}

bool msl::test::bitonic::adjustDirection(const DistributedArray<bool>& Dir, int i, bool d) {
	return odd(i) ? Dir.get(i - 1) : d;
}

template <class C> C msl::test::bitonic::min(C x, C y) {
	return (x > y) ? x : y;
}

template <class C> C msl::test::bitonic::max(C x, C y) {
	return (x <= y) ? x : y;
}

template <class C> C msl::test::bitonic::sorter(const DistributedArray<C>& A,
												 const DistributedArray<bool>& Dir, int i, C x) {
	typedef DistributedArray<C>& MutableA;

	if(odd(i))
		// elements at odd positions are processed by their neighbors
		return x;
	else {
		// A.get produces crashes in the current implementation
		C neighbor = A.get(i + 1);

		if(Dir.get(i)) {
			const_cast<MutableA>(A).set(i + 1, max(neighbor, x)); // slightly dirty 
			
			return min(neighbor, x);
		}
		else {
			const_cast<MutableA>(A).set(i + 1, min(neighbor, x)); // slightly dirty
			
			return max(neighbor, x);
		}
	}
}

int msl::test::bitonic::clsh(int k, int i) {
	int mask = 1 << k;
	
	return (i & mask) ? ((i ^ mask) << 1) + 1 : i << 1;
}

template<class C> msl::DistributedArray<C> msl::test::bitonic::bitonic(msl::DistributedArray<C> A) {
	int l, j;
	typedef C (*sorter_t)(const msl::DistributedArray<C>&, const msl::DistributedArray<bool>&, int, C);
	
	msl::DistributedArray<bool> SortDirection(n, &initDirection);
	A.mapIndexInPlace(curry((sorter_t)sorter<C>)(A)(SortDirection));

	for(l = 1; l < log2size; l++) {
		for(j = 0; j < log2size - l - 1; j++) {
			A.permute(curry(clsh)(log2size - 1));
			SortDirection.permute(curry(clsh)(log2size - 1));
		}

		for(j = log2size - l - 1; j < log2size; j++) {
			A.permute(curry(clsh)(log2size - 1));
			SortDirection.permute(curry(clsh)(log2size - 1));
			A.mapIndexInPlace(curry((sorter_t)sorter<C>)(A)(SortDirection));
			SortDirection.mapIndexInPlace(curry(adjustDirection)(SortDirection));
		}
	}

	return A;
}

//********************************************************************************
// Test: bitonic sort (algorithm similar to Quinn: Parallel Computing, p. 268)
//
// - n determines the size of the distributed array to sort.
// - nt determines the number of processor cores to use.
//
// ATTENTION: THE CURRENT IMPLEMENTATION OF THE BITONIC SORT ALGORITHM REQUIRES
// THE OLD IMPLEMENETATION OF THE GET-FUNCTION OF THE DISTRIBUTED ARRAY. IN THE
// CURRENT IMPLEMENTATION, THE BITONIC TEST SUITE PRODUCES PROGRAM CRASHES!!!
// THE OLD IMPLEMENTATION OF THE GET-FUNCTION IS COMMENTED OUT IN THE SOURCE FILE
// OF THE CLASS DISTRIBUTEDARRAY.
//********************************************************************************
void msl::test::bitonic::testBitonic(int n, int nt, bool showArrays) {
	msl::utl::printv("Welcome to the test suite for the bitonic sort algorithm.\n");

	Muesli::MSL_ARG1 = n;
	Muesli::MSL_ARG2 = nt;

	np = Muesli::MSL_numOfTotalProcs;

	log2size = log2(n);
	srand(Muesli::MSL_myId);
	msl::oal::setNumThreads(nt);

	msl::DistributedArray<int> A(n, &random);

	if(showArrays) {
		A.show();
	}

	A.debug(true);

	msl::DistributedArray<int> B = bitonic(A);
	
	if(showArrays) {
		B.show();
	}

	msl::utl::printv("The test has been successfully passed!\n\n");
}

int main(int argc, char** argv) {
	msl::InitSkeletons(argc, argv);
	msl::test::bitonic::testBitonic(4, 5, true);
	msl::TerminateSkeletons();
	return 0;
}
