// **********************************************************************************************************
// *																										*
// * This Software is published under the MIT License.														*
// * For more information and the license text, please refer to the file 'License' located in the zip file. *
// *																										*
// * Copyright (c) 2001-2011 Steffen Ernsting, Philipp Ciechanowicz, Michael Poldner and Herbert Kuchen     *                 
// *        		{s.ernsting|ciechanowicz|poldner|kuchen}@uni-muenster.de              					*
// *																										*	
// **********************************************************************************************************

#include "BBFrame.h"
#include "BBFrameWorkpool.h"
#include "BBProblemTracker.h"
#include "BBSolver.h"
#include "BranchAndBound.h"
#include "Double.h"
#include "TestBranchAndBound.h"

void msl::test::bab::testBranchAndBound(bool flag) {
	printf("Welcome to the test suite for all Branch & Bound related classes.\n\n");

	testBranchAndBoundFrame();
	testBranchAndBoundFrameWorkpool();
	//testBranchAndBoundProblemTracker(); //produces errors
	testBranchAndBoundSolver();
	testBranchAndBound();

	printf("\nAll tests have been successfully passed!\n\n");
}

void msl::test::bab::testBranchAndBoundFrame() {
	printf("Testing BBFrame...          ");

	int myId   = -1;
	int snd    = -1;
	int numSub = -1;
	int size;
	char* buffer;
	msl::test::Double x(1);
	msl::test::Double y(2);

	// testing constructors
	msl::BBFrame<msl::test::Double> a;
	msl::BBFrame<msl::test::Double> b(myId, &a, snd, numSub, &x);
	msl::BBFrame<msl::test::Double>* c = new msl::BBFrame<msl::test::Double>();

	// testing reduce, expand, and getSize
	size = b.getSize();
	buffer = new char[size];
	b.reduce(buffer, size);
	a.expand(buffer, size);

	// testing miscellaneous
	b.getData();
	b.getID();
	b.getNumOfSolvedSubProblems();
	b.getNumOfSubProblems();
	b.getOriginator();
	b.getParentProblem();
	b.setData(&y);
	b.setID(-2);
	b.setNumOfSolvedSubProblems(-2);
	b.setNumOfSubProblems(-2);
	b.setOriginator(-2);
	b.setParentProblem(c);

	// testing destructor
	delete c;

	printf("passed!\n");
}

void msl::test::bab::testBranchAndBoundFrameWorkpool() {
	printf("Testing BBFrameWorkpool...\n");
	printf("  ");

	msl::BBFrameWorkpool<msl::test::Double>* a = new msl::BBFrameWorkpool<msl::test::Double>(less);
	msl::BBFrame<Double> b;

	a->fresh();
	// calling a->get() on an empty workpool will result in an
	// emptyworkpoolexception -> insert element (bbframe) first
	a->insert(&b);
	a->get();
	a->getAverageLength();
	a->getBetterThanFunction();
	a->getMaxLength();
	a->isEmpty();
	a->reset();
	a->show(0);
	// calling a->get() on an empty workpool will result in an
	// emptyworkpoolexception -> insert element (bbframe) first
	a->insert(&b);
	a->top();

	delete a;

	printf("                            passed!\n");
}

void msl::test::bab::testBranchAndBoundProblemTracker() {
	printf("Testing BBProblemTracker... ");

	msl::BBProblemTracker<msl::test::Double>* a = new msl::BBProblemTracker<msl::test::Double>(1);
	msl::BBFrame<msl::test::Double> b;

	b.setData(new msl::test::Double(0));

	a->addProblem(&b);
	a->getProblemTrackerAverageLength();
	a->getProblemTrackerLength();
	a->getProblemTrackerMaxLength();
	a->getSolvedQueueLength();
	a->getSolvedQueueMaxLength();
	a->isSolvedQueueEmpty();
	a->isTrackerEmpty();
	a->problemSolved(&b);
	// reading/removing elements from an empty solved queue will result
	// in an emptyqueueexception -> insert element (bbfram) first
	a->writeToSolvedQueue(&b);
	a->readFromSolvedQueue();
	a->removeFromSolvedQueue();

	delete a;

	printf("passed!\n");
}

void msl::test::bab::testBranchAndBoundSolver() {
	printf("Testing BBSolver...\n");
	printf("  ");

	int length = 2;
	int numSub = 1;
	int n      = 1;
	msl::BBSolver<msl::test::Double>* a = new msl::BBSolver<msl::test::Double>(branch,
		bound, less, isSolution, getLowerBound, numSub, n);
	msl::BBSolver<msl::test::Double>** b = new msl::BBSolver<msl::test::Double>*[length];
	msl::ProcessorNo c = 1;
	msl::ProcessorNo d = 2;

	b[0] = new msl::BBSolver<msl::test::Double>(branch,
		bound, less, isSolution, getLowerBound, numSub, n);
	b[1] = new msl::BBSolver<msl::test::Double>(branch,
		bound, less, isSolution, getLowerBound, numSub, n);

	a->copy();
	a->setWorkmates(b, length, 0);
	a->show();
	
	// a->start() is not tested, since it produces MPI errors
	//a->start();
	printf("  note: start() is not tested\n");
	
	delete a;

	printf("                            passed!\n");
}

void msl::test::bab::testBranchAndBound() {
	printf("Testing BranchAndBound...\n");

	int numSub = 1;
	int n      = 1;
	msl::BBSolver<msl::test::Double> x(branch, bound, less, isSolution, getLowerBound, numSub, n);
	msl::BranchAndBound<msl::test::Double> b(x, n);
	msl::BranchAndBound<msl::test::Double> c(branch, bound, less, isSolution, getLowerBound, numSub, n);
	msl::ProcessorNo d = 1;
	msl::ProcessorNo e = 2;

	// test branch and bound

	b.copy();
	b.setPredecessors(&d, 1);
	b.setSuccessors(&e, 1);
	b.show();

	// b.start() is not tested, since it produces MPI errors
	//b.start();
	printf("  note: start() is not tested\n");

	printf("                            passed!\n");
}

// auxiliary functions

void msl::test::bab::bound(msl::test::Double* a) {
}

msl::test::Double** msl::test::bab::branch(msl::test::Double* a, int* b) {
	return &a;
}

int msl::test::bab::getLowerBound(msl::test::Double* a) {
	return 0;
}

bool msl::test::bab::isSolution(msl::test::Double* a) {
	return true;
}

bool msl::test::bab::less(msl::test::Double* a, msl::test::Double* b) {
	return true;
}


int main(int argc, char** argv) {
	msl::InitSkeletons(argc, argv);
	msl::test::bab::testBranchAndBound(true);
	msl::TerminateSkeletons();
	return 0;
}
