// **********************************************************************************************************
// *																										*
// * This Software is published under the MIT License.														*
// * For more information and the license text, please refer to the file 'License' located in the zip file. *
// *																										*
// * Copyright (c) 2001-2011 Steffen Ernsting, Philipp Ciechanowicz, Michael Poldner and Herbert Kuchen     *                 
// *        		{s.ernsting|ciechanowicz|poldner|kuchen}@uni-muenster.de              					*
// *																										*	
// **********************************************************************************************************

#include <cstdarg>
#include <cstdio>

#include "Curry.h"
#include "DistributedMatrix.h"
#include "Muesli.h"
#include "TestDistributedMatrix.h"
#include "Utility.h"

// test functions

void msl::test::dm::testDistributedMatrix() {
	msl::utl::printv("Welcome to the test suite for the DistributedMatrix.\n");

	if(Muesli::MSL_numOfTotalProcs == PROCESSORS) {
		msl::utl::printv("Note that only the first process prints its output!\n");
		msl::utl::printv("The test parameters are as follows:\n");
		msl::utl::printv("n       = %d\n", N);
		msl::utl::printv("m       = %d\n", M);
		msl::utl::printv("nLocal  = %d\n", N_LOCAL);
		msl::utl::printv("mLocal  = %d\n", M_LOCAL);

		testAuxiliaryFunctions();
		testConstructors();
		testDestructor();
		testSkeletons();

		msl::utl::printv("\nall test passed!\n\n");
	}
	else {
		msl::utl::printv("Please restart this suite with np = %d processors!\n\n", PROCESSORS);
	}
}

void msl::test::dm::testAuxiliaryFunctions() {
	msl::DistributedMatrix<int> dm(N, M, init, N_LOCAL, M_LOCAL);
	msl::DistributedMatrix<int> dm2(dm);

	msl::utl::printv("\n");
	msl::utl::printv("****************************************\n");
	msl::utl::printv("* 1. testing auxiliary functions       *\n");
	msl::utl::printv("****************************************\n");
	
	msl::utl::printv("\ndm =\n");
	dm.show();
	msl::utl::printv("\n");

	for(int i = 0; i < N; i++) {
		for(int j = 0; j < M; j++) {
			msl::utl::printv("dm.get(%d, %d)\t\t= %d\n", i, j, dm.get(i, j));
		}
	}

	msl::utl::printv("dm.getBlocksInCol()\t= %d\n", dm.getBlocksInCol());
	msl::utl::printv("dm.getBlocksInRow()\t= %d\n", dm.getBlocksInRow());
	msl::utl::printv("dm.getCols()\t\t= %d\n", dm.getCols());
	msl::utl::printv("dm.getFirstCol()\t= %d\n", dm.getFirstCol());
	msl::utl::printv("dm.getFirstRow()\t= %d\n", dm.getFirstRow());
	msl::utl::printv("note: dm.getGlobalLocal(int, int) not tested\n");
	
	for(int i = 0; i < N_LOCAL; i++) {
		for(int j = 0; j < M_LOCAL; j++) {
			msl::utl::printv("dm.getLocal(%d, %d)\t= %d\n", i, j, dm.getLocal(i, j));
		}
	}

	msl::utl::printv("dm.getLocalCols()\t= %d\n", dm.getLocalCols());
	msl::utl::printv("note: dm.getLocalGlobal(int, int) not tested\n");
	msl::utl::printv("dm.getLocalRows()\t= %d\n", dm.getLocalRows());
	msl::utl::printv("dm.getRows()\t\t= %d\n", dm.getRows());
	
	for(int i = -1; i < N + 1; i++) {
		for(int j = -1; j < M + 1; j++) {
			msl::utl::printv("dm.isLocal(%d, %d)\t= %d\n", i, j, dm.isLocal(i, j));
		}
	}
	
	for(int i = 0; i < N_LOCAL; i++) {
		for(int j = 0; j < M_LOCAL; j++) {
			dm.setLocal(i, j, i * N_LOCAL + j + 1);
			msl::utl::printv("dm.setLocal(%d, %d, %d)\n", i, j, i * N_LOCAL + j + 1);
		}
	}

	dm.show();
	msl::utl::printv("operator <<, dm =\n");
	std::cout << dm2;

	msl::utl::printv("\n\nall auxiliary function tests passed!\n");
}

void msl::test::dm::testConstructors() {
	int** b = new int*[N];

	for(int i = 0; i < N; i++) {
		b[i] = new int[M];

		for(int j = 0; j < M; j++) {
			b[i][j] = i * N + j + 10;
		}
	}

	msl::DistributedMatrix<int> dm1(N, M, N_LOCAL, M_LOCAL);
	msl::DistributedMatrix<int> dm2(N, M, INITIAL, N_LOCAL, M_LOCAL);
	msl::DistributedMatrix<int> dm3(N, M, b, N_LOCAL, M_LOCAL);
	msl::DistributedMatrix<int> dm4(N, M, init, N_LOCAL, M_LOCAL);
	msl::DistributedMatrix<int> dm5(N, M, curry(init), N_LOCAL, M_LOCAL);
	msl::DistributedMatrix<int> dm6(dm5);

	msl::utl::printv("\n");
	msl::utl::printv("****************************************\n");
	msl::utl::printv("* 2. testing constructors              *\n");
	msl::utl::printv("****************************************\n");
	msl::utl::printv("\ninitial = %d\n", INITIAL);
	msl::utl::printv("b       = [");

	for(int i = 0; i < N; i++) {
		for(int j = 0; j < M; j++) {
			msl::utl::printv("%d", b[i][j]);

			if(j < M - 1) {
				msl::utl::printv(" ");
			}
			else {
				msl::utl::printv("]\n");
			}
		}

		if(i < N - 1) {
			msl::utl::printv("          [");
		}
	}

	msl::utl::printv("\ndm1, all values uninitialized\n");
	dm1.show();
	msl::utl::printv("\ndm2, all values set to initial\n");
	dm2.show();
	msl::utl::printv("\ndm3, values initialized with b\n");
	dm3.show();
	msl::utl::printv("\ndm4, values initialized with f\n");
	dm4.show();
	msl::utl::printv("\ndm5, values initialized with curry(f)\n");
	dm5.show();
	msl::utl::printv("\ndm6, values initialized with dm5\n");
	dm6.show();
	msl::utl::printv("\nall constructor tests passed!\n");
}

void msl::test::dm::testDestructor() {
	msl::DistributedMatrix<int>* dm = new msl::DistributedMatrix<int>(N, M, init, N_LOCAL, M_LOCAL);

	msl::utl::printv("\n");
	msl::utl::printv("****************************************\n");
	msl::utl::printv("* 3. testing destructor                *\n");
	msl::utl::printv("****************************************\n");

	delete dm;

	msl::utl::printv("\ndestructor test passed!\n");
}

void msl::test::dm::testSkeletons() {
	msl::utl::printv("\n");
	msl::utl::printv("****************************************\n");
	msl::utl::printv("* 4. testing skeletons                 *\n");
	msl::utl::printv("****************************************\n");

	testSkeletonsComputation();
	testSkeletonsCommunication();

	msl::utl::printv("\nall skeleton tests passed!\n");
}

void msl::test::dm::testSkeletonsCommunication() {
	int** b = new int*[N];

	for(int i = 0; i < N; i++) {
		b[i] = new int[M];
	}

	msl::DistributedMatrix<int> dm1(N, M, init, N_LOCAL, M_LOCAL);
	msl::DistributedMatrix<int> dm2(dm1);
	msl::DistributedMatrix<int> dm3(dm1);
	msl::DistributedMatrix<int> dm4(dm1);
	msl::DistributedMatrix<int> dm5(dm1);
	msl::DistributedMatrix<int> dm6(dm1);

	msl::utl::printv("\n");
	msl::utl::printv("****************************************\n");
	msl::utl::printv("* 4.2 testing communication skeletons  *\n");
	msl::utl::printv("****************************************\n");
	
	msl::utl::printv("\ndm =\n");
	dm1.show();

	msl::utl::printv("\nbroadcast,          dm =\n");
	dm1.broadcast(1, 1);
	dm1.show();
	msl::utl::printv("\nbroadcastPartition, dm =\n");
	dm2.broadcastPartition(1, 1);
	dm2.show();
	msl::utl::printv("\ngather,             dm =\n");
	dm3.gather(b);

	for(int i = 0; i < N; i++) {
		msl::utl::printv("[");

		for(int j = 0; j < M; j++) {
			msl::utl::printv("%d", b[i][j]);

			if(j < M - 1) {
				msl::utl::printv(" ");
			}
			else {
				msl::utl::printv("]\n");
			}
		}
	}

	msl::utl::printv("\npermutePartition,   dm =\n");
	dm4.permutePartition(newRow, newCol);
	dm4.show();
	msl::utl::printv("\nrotateCols,         dm =\n");
	dm5.rotateCols(rotateCols);
	dm5.show();
	msl::utl::printv("\nrotateRows,         dm =\n");
	dm6.rotateRows(rotateRows);
	dm6.show();

	msl::utl::printv("\nall communication skeleton tests passed!\n");
}

void msl::test::dm::testSkeletonsComputation() {
	msl::utl::printv("\n");
	msl::utl::printv("****************************************\n");
	msl::utl::printv("* 4.1 testing computation skeletons    *\n");
	msl::utl::printv("****************************************\n");

	testSkeletonsMap();
	testSkeletonsZip();
	testSkeletonsRemaining();

	msl::utl::printv("\nall computation skeleton tests passed!\n");
}

void msl::test::dm::testSkeletonsMap() {
	msl::DistributedMatrix<int> dm1(N, M, init, N_LOCAL, M_LOCAL);
	msl::DistributedMatrix<int> dm2(dm1);
	msl::DistributedMatrix<int> dm3(dm1);
	msl::DistributedMatrix<int> dm4(dm1);
	msl::DistributedMatrix<int> dm5(dm1);

	msl::utl::printv("\n");
	msl::utl::printv("****************************************\n");
	msl::utl::printv("* 4.1.1 testing map                    *\n");
	msl::utl::printv("****************************************\n");

	msl::utl::printv("\ndm =\n");
	dm1.show();
	msl::utl::printv("\nmap,             dm =\n");
	dm2.map(increase).show();
	msl::utl::printv("\nmapIndex,        dm =\n");
	dm3.mapIndex(increaseIndex).show();
	msl::utl::printv("\nmapIndexInPlace, dm =\n");
	dm4.mapIndexInPlace(increaseIndex);
	dm4.show();
	msl::utl::printv("\nmapInPlace,      dm =\n");
	dm5.mapInPlace(increase);
	dm5.show();
	msl::utl::printv("\nnote: mapPartitionInPlace not tested\n");

	msl::utl::printv("\nall map skeleton tests passed!\n");
}

void msl::test::dm::testSkeletonsRemaining() {
	msl::DistributedMatrix<int> dm(N, M, init, N_LOCAL, M_LOCAL);

	msl::utl::printv("\n");
	msl::utl::printv("****************************************\n");
	msl::utl::printv("* 4.1.3 testing remaining skeletons    *\n");
	msl::utl::printv("****************************************\n");

	msl::utl::printv("\ndm =\n");
	dm.show();

	msl::utl::printv("\nfold                                = %d\n", dm.fold(add));
	/*msl::utl::printv("mapIndexInPlaceFold                 = %d\n", dm.mapIndexInPlaceFold(increaseIndex, add));
	msl::utl::printv("mapIndexInPlacePermutePartition, dm = ");
	dm.mapIndexInPlacePermutePartition(increaseIndex, leftShiftPartition);
	dm.show();
	msl::utl::printv("mapIndexInPlaceScan,             dm = ");
	dm.mapIndexInPlaceScan(increaseIndex, add);
	dm.show();*/

	msl::utl::printv("\nall remaining skeleton tests passed!\n");
}

void msl::test::dm::testSkeletonsZip() {
	msl::DistributedMatrix<int> dm1(N, M, init, N_LOCAL, M_LOCAL);
	msl::DistributedMatrix<int> dm2(dm1);

	msl::utl::printv("\n");
	msl::utl::printv("****************************************\n");
	msl::utl::printv("* 4.1.2 testing zip                    *\n");
	msl::utl::printv("****************************************\n");

	msl::utl::printv("\ndm1 =\n");
	dm1.show();
	msl::utl::printv("\ndm2 =\n");
	dm2.show();

	msl::utl::printv("\nzip,                 dm =\n");
	dm1.copy().zipWith(dm2.copy(), add).show();
	msl::utl::printv("\nzipWithIndex,        dm =\n");
	dm1.copy().zipWithIndex(dm2.copy(), addIndex).show();
	msl::utl::printv("\nzipWithInPlace,      dm =\n");
	dm1.zipWithInPlace(dm2.copy(), add);
	dm1.show();
	msl::utl::printv("\nzipWithIndexInPlace, dm =\n");
	dm2.zipWithIndexInPlace(dm2.copy(), addIndex);
	dm2.show();

	msl::utl::printv("\nall zip skeleton tests passed!\n");
}

// argument functions

int msl::test::dm::add(int x, int y) {
	return x + y;
}

int msl::test::dm::addIndex(int rowIndex, int colIndex, int value1, int value2) {
	return value1 + value2 + rowIndex + colIndex;
}

int msl::test::dm::increase(int value) {
	return value + 1;
}

int msl::test::dm::increaseIndex(int rowIndex, int colIndex, int value) {
	return value + rowIndex + colIndex;
}

int msl::test::dm::init(int rowIndex, int colIndex) {
	return rowIndex * N + colIndex + 20;
}

int msl::test::dm::newCol(int rowIndex, int colIndex) {
	return (colIndex + 1) % N_LOCAL;
}

int msl::test::dm::newRow(int rowIndex, int colIndex) {
	return (rowIndex + 1) % M_LOCAL;
}

int msl::test::dm::rotateCols(int colIndex) {
	return (colIndex + 1) % M;
}

int msl::test::dm::rotateRows(int rowIndex) {
	return (rowIndex + 1) % N;
}


int main(int argc, char** argv) {
	msl::InitSkeletons(argc, argv);
	msl::test::dm::testDistributedMatrix();
	msl::TerminateSkeletons();
	return 0;
}
