// **********************************************************************************************************
// *																										*
// * This Software is published under the MIT License.														*
// * For more information and the license text, please refer to the file 'License' located in the zip file. *
// *																										*
// * Copyright (c) 2001-2011 Steffen Ernsting, Philipp Ciechanowicz, Michael Poldner and Herbert Kuchen     *                 
// *        		{s.ernsting|ciechanowicz|poldner|kuchen}@uni-muenster.de              					*
// *																										*	
// **********************************************************************************************************

#include <cstdarg>
#include <cstdio>

#include "BsrSubmatrix.h"
#include "CrsSubmatrix.h"
#include "Curry.h"
#include "DistributedSparseMatrix.h"
#include "Muesli.h"
#include "TestDistributedSparseMatrix.h"
#include "Utility.h"

// some variables

msl::DistributedSparseMatrix<int>* dsmA;
msl::DistributedSparseMatrix<int>* dsmB;
int** a;
int** b;

// test functions

void msl::test::dsm::testDistributedSparseMatrix() {
	msl::utl::printv("Welcome to the test suite for the DistributedSparseMatrix.\n");

	if(Muesli::MSL_numOfTotalProcs == PROCESSORS) {
		msl::utl::printv("Note that only the first process prints its output!\n");
		msl::utl::printv("The test parameters are as follows:\n");
		msl::utl::printv("n = %d\n", N);
		msl::utl::printv("m = %d\n", M);
		msl::utl::printv("r = %d\n", R);
		msl::utl::printv("c = %d\n", C);

		init();
		testAuxiliaryFunctions();
		testConstructors();
		testDestructor();
		testSkeletons();

		msl::utl::printv("\nall test passed!\n\n");
	}
	else {
		msl::utl::printv("Please restart this suite with np = %d processors!\n\n", PROCESSORS);
	}
}

// argument functions

int msl::test::dsm::add(int a, int b) {
	return a + b;
}

int msl::test::dsm::addIndex(int a, int b, int rowIndex, int colIndex) {
	int result = a + b;

	if(rowIndex == -1 && colIndex == -1) {
		return result;
	}
	else {
		return result + rowIndex + colIndex;
	}
}

bool msl::test::dsm::count(int value) {
	if(value == 0) {
		return false;
	}
	else {
		return true;
	}
}

bool msl::test::dsm::elementCount(int value) {
	return true;
}

bool msl::test::dsm::elementCountIndex(int value, int rowIndex, int colIndex) {
	if(rowIndex > 1 && colIndex > 1) {
		return true;
	}
	else {
		return false;
	}
}

int msl::test::dsm::filter(int value) {
	if(value > 1) {
		return value;
	}
	else {
		return 0;
	}
}

int msl::test::dsm::filterIndex(int value, int rowIndex, int colIndex) {
	if(value > 1 && rowIndex > 1 && colIndex > 1) {
		return value;
	}
	else {
		return 0;
	}
}

int msl::test::dsm::increase(int a) {
	return a + 1;
}

int msl::test::dsm::increaseIndex(int a, int rowIndex, int colIndex) {
	return a + rowIndex + colIndex;
}

int msl::test::dsm::rotateColumns(int columnIndex) {
	return -1;
}

int msl::test::dsm::rotateRows(int rowIndex) {
	return 1;
}

// various functions

void msl::test::dsm::init() {
	a = new int*[N];
	b = new int*[N];

	for(int i = 0; i < N; i++) {
		a[i] = new int[M];
		b[i] = new int[M];

		for(int j = 0; j < M; j++) {
			a[i][j] = 0;
			b[i][j] = 0;
		}
	}

	a[1][1] = 1;
	a[2][2] = 2;
	a[3][0] = 3;
	a[3][3] = 4;

	b[0][0] = 1;
	b[1][1] = 2;
	b[0][3] = 3;
	b[3][0] = 4;

	dsmA = new msl::DistributedSparseMatrix<int>(N, M, R, C, 0, a);
	dsmB = new msl::DistributedSparseMatrix<int>(N, M, R, C, 0, b);
}

// test functions

void msl::test::dsm::testAuxiliaryFunctions() {
	int* x = new int[N];
	int* y = new int[M];
	msl::DistributedSparseMatrix<int> dsm(*dsmA);

	for(int i = 0; i < N; i++) {
		x[i] = 0;
	}

	for(int i = 0; i < M; i++) {
		y[i] = 0;
	}

	msl::utl::printv("\n");
	msl::utl::printv("****************************************\n");
	msl::utl::printv("* 1. testing auxiliary functions       *\n");
	msl::utl::printv("****************************************\n");
	
	msl::utl::printv("\ndsm =\n");
	dsm.print();

	msl::utl::printv("\nNote: addSubmatrix(S*) not tested.");
	msl::utl::printv("\ngetC()                 = %d", dsm.getC());
	msl::utl::printv("\ngetColumn(3)           = [");
	dsm.getColumn(3, x);

	for(int i = 0; i < N; i++) {
		msl::utl::printv("%d", x[i]);

		if(i < N - 1) {
			msl::utl::printv(" ");
		}
		else {
			msl::utl::printv("]");
		}
	}

	msl::utl::printv("\ngetColumnCount()       = %d", dsm.getColumnCount());
	msl::utl::printv("\ngetElement(2, 2)       = %d", dsm.getElement(2, 2));
	msl::utl::printv("\ngetElementCount()      = %d", dsm.getElementCount(&count));
	msl::utl::printv("\ngetElementCountIndex() = %d", dsm.getElementCountIndex(elementCountIndex));
	msl::utl::printv("\ngetElementLocal(1, 1)  = %d", dsm.getElementLocal(1, 1));
	msl::utl::printv("\ngetR()                 = %d", dsm.getR());
	msl::utl::printv("\ngetRow(3)              = [");
	dsm.getRow(3, y);

	for(int i = 0; i < M; i++) {
		msl::utl::printv("%d", y[i]);

		if(i < M - 1) {
			msl::utl::printv(" ");
		}
		else {
			msl::utl::printv("]");
		}
	}

	msl::utl::printv("\ngetRowCount()          = %d", dsm.getRowCount());
	msl::utl::printv("\ntesting getSubmatrixById()");
	msl::utl::printv("\ntesting getSubmatrixByIndex()");
	msl::utl::printv("\ngetSubmatrixCount()    = %d", dsm.getSubmatrixCount());
	msl::utl::printv("\nisStoredLocally(0)     = %d", dsm.isStoredLocally(0));
	msl::utl::printv("\ntesting multiply(y, x)\n  y = [");
	
	for(int i = 0; i < N; i++) {
		x[i] = 0;
	}

	for(int i = 0; i < M; i++) {
		y[i] = 1;
	}

	dsm.multiply(y, x);

	for(int i = 0; i < M; i++) {
		msl::utl::printv("%d", y[i]);

		if(i < M - 1) {
			msl::utl::printv(" ");
		}
		else {
			msl::utl::printv("]\n  x = [");
		}
	}

	for(int i = 0; i < N; i++) {
		msl::utl::printv("%d", x[i]);

		if(i < N - 1) {
			msl::utl::printv(" ");
		}
		else {
			msl::utl::printv("]");
		}
	}

	msl::utl::printv("\ntesting pack()");
	dsm.pack();
	msl::utl::printv("\ntesting packSubmatrices()");
	dsm.packSubmatrices();
	msl::utl::printv("\ntesting msl::utl::printv()\n");
	dsm.print();
	msl::utl::printv("\ntesting removeEmptySubmatrices()");
	dsm.removeEmptySubmatrices();
	msl::utl::printv("\ntesting rotateColumn(), dsm =\n");
	dsm.rotateColumn(0, -1);
	dsm.print();
	dsm.rotateColumn(0, 1);
	msl::utl::printv("\ntesting rotateColumns(), dsm =\n");
	dsm.rotateColumns(rotateColumns);
	dsm.print();
	dsm.rotateColumns(rotateRows);
	msl::utl::printv("\ntesting rotateRow(), dsm =\n");
	dsm.rotateRow(3, -1);
	dsm.print();
	dsm.rotateRow(3, 1);
	msl::utl::printv("\ntesting rotateRows(), dsm =\n");
	dsm.rotateRows(rotateColumns);
	dsm.print();
	dsm.rotateRows(rotateRows);
	msl::utl::printv("\ntesting setElement(), dsm =\n");
	dsm.setElement(1, 0, 0);
	dsm.print();
	dsm.setElement(0, 0, 0);
	msl::utl::printv("\ntesting setPresicion()");
	dsm.setPrecision(2);

	msl::utl::printv("\n\nall auxiliary function tests passed!\n");
}

void msl::test::dsm::testConstructors() {
	msl::utl::printv("\n");
	msl::utl::printv("****************************************\n");
	msl::utl::printv("* 2. testing constructors              *\n");
	msl::utl::printv("****************************************\n");
	msl::utl::printv("\n");
	msl::utl::printv("b =\n");

	for(int i = 0; i < N; i++) {
		msl::utl::printv("[");

		for(int j = 0; j < M; j++) {
			msl::utl::printv("%d", b[i][j]);

			if(j < M - 1) {
				msl::utl::printv(" ");
			}
			else {
				msl::utl::printv("]");
			}
		}

		msl::utl::printv("\n");
	}

	msl::utl::printv("\ndsm1, all values uninitialized\n");
	msl::DistributedSparseMatrix<int> dsm1(N, M, R, C, 0);
	dsm1.print();
	msl::utl::printv("\ndsm2, values initialized with b\n");
	msl::DistributedSparseMatrix<int> dsm2(N, M, R, C, 0, b);
	dsm2.print();
	msl::utl::printv("\ndsm3, values initialized with dsm2\n");
	msl::DistributedSparseMatrix<int> dsm3(dsm2);
	dsm3.print();
	
	msl::utl::printv("\nall constructor tests passed!\n");
}

void msl::test::dsm::testDestructor() {
	msl::utl::printv("\n");
	msl::utl::printv("****************************************\n");
	msl::utl::printv("* 3. testing destructor                *\n");
	msl::utl::printv("****************************************\n");

	msl::DistributedSparseMatrix<int>* dsm = new msl::DistributedSparseMatrix<int>(N, M, R, C, 0);
	delete dsm;

	msl::utl::printv("\ndestructor test passed!\n");
}

void msl::test::dsm::testSkeletons() {
	msl::utl::printv("\n");
	msl::utl::printv("****************************************\n");
	msl::utl::printv("* 4. testing skeletons                 *\n");
	msl::utl::printv("****************************************\n");

	testSkeletonsComputation();
	//testSkeletonsCommunication();

	msl::utl::printv("\nall skeleton tests passed!\n");
}

void msl::test::dsm::testSkeletonsCommunication() {
	// currently, the class DSM does not implement any pure communication skeletons
}

void msl::test::dsm::testSkeletonsComputation() {
	msl::utl::printv("\n");
	msl::utl::printv("****************************************\n");
	msl::utl::printv("* 4.1 testing computation skeletons    *\n");
	msl::utl::printv("****************************************\n");

	testSkeletonsMap();
	testSkeletonsZip();
	testSkeletonsRemaining();

	msl::utl::printv("\nall computation skeleton tests passed!\n");
}

void msl::test::dsm::testSkeletonsMap() {
	msl::DistributedSparseMatrix<int> dsm1(*dsmA);
	msl::DistributedSparseMatrix<int> dsm2(*dsmA);

	msl::utl::printv("\n");
	msl::utl::printv("****************************************\n");
	msl::utl::printv("* 4.1.1 testing map                    *\n");
	msl::utl::printv("****************************************\n");

	msl::utl::printv("\ndsm =\n");
	dsm1.print();
	msl::utl::printv("\nmap,             dsm =\n");
	dsm1.map(increase)->print();
	msl::utl::printv("\nmapIndex,        dsm =\n");
	dsm1.mapIndex(increaseIndex)->print();
	msl::utl::printv("\nmapInPlace,      dsm =\n");
	dsm1.mapInPlace(increase);
	dsm1.print();
	msl::utl::printv("\nmapIndexInPlace, dsm =\n");
	dsm2.mapIndexInPlace(increaseIndex);
	dsm2.print();

	msl::utl::printv("\nall map skeleton tests passed!\n");
}

void msl::test::dsm::testSkeletonsRemaining() {
	int* a = new int[N * M];

	msl::DistributedSparseMatrix<int> dsm(*dsmA);

	for(int i = 0; i < N * M; i++) {
		a[i] = 0;
	}

	msl::utl::printv("\n");
	msl::utl::printv("****************************************\n");
	msl::utl::printv("* 4.1.3 testing remaining skeletons    *\n");
	msl::utl::printv("****************************************\n");

	msl::utl::printv("\ndsm =\n");
	dsm.print();
	msl::utl::printv("\nfilter            = [");
	dsm.filter(filter, a);

	for(int i = 0; i < N * M; i++) {
		msl::utl::printv("%d", a[i]);

		a[i] = 0;

		if(i < N * M - 1) {
			msl::utl::printv(" ");
		}
		else {
			msl::utl::printv("]\n");
		}
	}

	msl::utl::printv("\nfilterIndex       = [");
	dsm.filterIndex(filterIndex, a);

	for(int i = 0; i < N * M; i++) {
		msl::utl::printv("%d", a[i]);

		a[i] = 0;

		if(i < N * M - 1) {
			msl::utl::printv(" ");
		}
		else {
			msl::utl::printv("]\n");
		}
	}

	msl::utl::printv("\nfilterIndexRow    = [");
	dsm.filterIndexRow(filterIndex, a, 2);

	for(int i = 0; i < N * M; i++) {
		msl::utl::printv("%d", a[i]);

		a[i] = 0;

		if(i < N * M - 1) {
			msl::utl::printv(" ");
		}
		else {
			msl::utl::printv("]\n");
		}
	}

	msl::utl::printv("\nfilterIndexColumn = [");
	dsm.filterIndexColumn(filterIndex, a, 3);

	for(int i = 0; i < N * M; i++) {
		msl::utl::printv("%d", a[i]);

		a[i] = 0;

		if(i < N * M - 1) {
			msl::utl::printv(" ");
		}
		else {
			msl::utl::printv("]\n");
		}
	}

	msl::utl::printv("\nfold              = %d\n", dsm.fold(add));
	msl::utl::printv("\nfoldIndex         = %d\n", dsm.foldIndex(addIndex));

	msl::utl::printv("\nall remaining skeleton tests passed!\n");
}

void msl::test::dsm::testSkeletonsZip() {
	msl::DistributedSparseMatrix<int> dsm1(*dsmA);
	msl::DistributedSparseMatrix<int> dsm2(*dsmB);
	msl::DistributedSparseMatrix<int> dsm3(*dsmA);

	msl::utl::printv("\n");
	msl::utl::printv("****************************************\n");
	msl::utl::printv("* 4.1.2 testing zip                    *\n");
	msl::utl::printv("****************************************\n");

	msl::utl::printv("\ndsm1 =\n");
	dsm1.print();
	msl::utl::printv("\ndsm2 =\n");
	dsm2.print();

	msl::utl::printv("\ndsm1.zip(dsm2)             =\n");
	dsm1.zip(dsm2, add)->print();
	msl::utl::printv("\ndsm1.zipIndex(dsm2)        =\n");
	dsm1.zipIndex(dsm2, addIndex)->print();
	msl::utl::printv("\ndsm1.zipInPlace(dsm2)      =\n");
	dsm1.zipInPlace(dsm2, add);
	dsm1.print();
	msl::utl::printv("\ndsm1.zipIndexInPlace(dsm2) =\n");
	dsm3.zipIndexInPlace(dsm2, addIndex);
	dsm3.print();

	msl::utl::printv("\nall zip skeleton tests passed!\n");
}


int main(int argc, char** argv) {
	msl::InitSkeletons(argc, argv);
	msl::test::dsm::testDistributedSparseMatrix();
	msl::TerminateSkeletons();
	return 0;
}
