// **********************************************************************************************************
// *																										*
// * This Software is published under the MIT License.														*
// * For more information and the license text, please refer to the file 'License' located in the zip file. *
// *																										*
// * Copyright (c) 2001-2011 Steffen Ernsting, Philipp Ciechanowicz, Michael Poldner and Herbert Kuchen     *                 
// *        		{s.ernsting|ciechanowicz|poldner|kuchen}@uni-muenster.de              					*
// *																										*	
// **********************************************************************************************************

#include <cmath>

#include "Array.h"
#include "DCStreamSolver.h"
#include "Final.h"
#include "Initial.h"
#include "Muesli.h"
#include "Pipe.h"
#include "StreamDC.h"
#include "TestDivideAndConquer.h"
#include "Utility.h"

/* This class is supposed to test the functionality of the stream version of the
   Divide & Conquer skeleton by implementing a simple sorting algorithm. The
   numbers to sort are stored in a container of type msl::Array<T>, which is
   merely an array being aware of its size. The D&C skeleton is used with the
   usual initial-final-construct inside a pipe. While the sorting is done
   correctly, the program eventually crashes when executing either the function
   WorkpoolManager.isEmpty() or SolutionpoolManager2.hasSolution(). I am not
   sure why this happens, since I did not implement any of the classes. It is
   possible that I am using the skeleton in a wrong way by passing wrong
   parameters to the classes DCStreamSolver and/or StreamDC. Thus, the benchmark
   on the one hand computes the correct result, but unfortunately crashes when
   doing some clean-up.
*/

/* Main function.
*/
void msl::test::dac::testDivideAndConquer() {
	if(Muesli::MSL_numOfTotalProcs != 3) {
		printf("Please restart the program with np = 3 processes!\n");
	}

	// number of processes of the d&c skeleton
	int n = 1;
	// number of master solvers
	int m = 1;
	msl::Initial<Array<int> > in(initial);
	msl::DCStreamSolver<msl::test::Array<int>, msl::test::Array<int> > dcss(divide,
		combine, solve, isSimple, D, m, n);
	msl::StreamDC<msl::test::Array<int>, msl::test::Array<int> > sdc(dcss, 1, 1);
	msl::Final<msl::test::Array<int> > fi(final);
	msl::Pipe pi(in, sdc, fi);

	pi.start();
}

/* Create array to sort.
*/
msl::test::Array<int>* msl::test::dac::initial(msl::Empty) {
	int n;
	msl::test::Array<int>* a;

	if(flag) {
		flag = false;

		n = 8;
		a = new msl::test::Array<int>(n);

		a->setValue(0, 4);
		a->setValue(1, 2);
		a->setValue(2, 7);
		a->setValue(3, 1);
		a->setValue(4, 6);
		a->setValue(5, 8);
		a->setValue(6, 5);
		a->setValue(7, 3);

		printf("problem:  ");
		a->print();

		return a;
	}
	else {
		return NULL;
	}
}

/* Determine, whether the given array stores two or less elements.
*/
bool msl::test::dac::isSimple(msl::test::Array<int>* problem) {
	if(problem->getLength() <= 2) {
		return true;
	}
	else {
		return false;
	}
}

/* Split an array of size three or larger into two subproblems.
*/
msl::test::Array<int>** msl::test::dac::divide(msl::test::Array<int>* problem) {
	int n  = problem->getLength();
	int nx = (int)floor((double)n / D);
	int ny = (int)ceil((double)n / D);
	msl::test::Array<int>** result = new msl::test::Array<int>*[D];
	msl::test::Array<int>* x = new msl::test::Array<int>(nx);
	msl::test::Array<int>* y = new msl::test::Array<int>(ny);

	for(int i = 0; i < nx; i++) {
		x->setValue(i, problem->getValue(i));
	}

	for(int i = 0; i < ny; i++) {
		y->setValue(i, problem->getValue(i + nx));
	}

	result[0] = x;
	result[1] = y;

	return result;
}

/* Sort an array of size two or less.
*/
msl::test::Array<int>* msl::test::dac::solve(msl::test::Array<int>* problem) {
	int n = problem->getLength();
	int tmp;
	int x0;
	int x1;

	if(n == 1) {
		return problem;
	}
	// n should be 2
	else {
		x0 = problem->getValue(0);
		x1 = problem->getValue(1);

		if(x0 <= x1) {
			return problem;
		}
		else {
			tmp = x0;
			problem->setValue(0, x1);
			problem->setValue(1, tmp);

			return problem;
		}
	}
}

/* Merge two sorted arrays.
*/
msl::test::Array<int>* msl::test::dac::combine(msl::test::Array<int>** solution) {
	msl::test::Array<int>* x = solution[0];
	msl::test::Array<int>* y = solution[1];
	int ix = 0;
	int iy = 0;
	int nx = x->getLength();
	int ny = y->getLength();
	int n  = nx + ny;
	msl::test::Array<int>* result = new msl::test::Array<int>(n);

	for(int i = 0; i < n; i++) {
		if(ix == nx) {
			result->setValue(i, y->getValue(iy++));
		}
		else if(iy == ny) {
			result->setValue(i, x->getValue(ix++));
		}
		else if(x->getValue(ix) < y->getValue(iy)) {
			result->setValue(i, x->getValue(ix++));
		}
		else {
			result->setValue(i, y->getValue(iy++));
		}
	}

	return result;
}

void msl::test::dac::final(msl::test::Array<int>* solution) {
	printf("solution: ");
	solution->print();
	printf("ATTENTION: Unfortunately, the program crashes when doing some clean-up\n");
	printf("work!. For details, please refer to the file 'TestDivideAndConquer.cpp'\n");
}



int main(int argc, char** argv) {
	msl::InitSkeletons(argc, argv);
	msl::test::dac::testDivideAndConquer();
	msl::TerminateSkeletons();
	return 0;
}

