// **********************************************************************************************************
// *																										*
// * This Software is published under the MIT License.														*
// * For more information and the license text, please refer to the file 'License' located in the zip file. *
// *																										*
// * Copyright (c) 2001-2011 Steffen Ernsting, Philipp Ciechanowicz, Michael Poldner and Herbert Kuchen     *                 
// *        		{s.ernsting|ciechanowicz|poldner|kuchen}@uni-muenster.de              					*
// *																										*	
// **********************************************************************************************************

#include <math.h>
#include <iostream>

#include "DistributedArray.h"
#include "Muesli.h"
#include "OAL.h"
#include "TestFft.h"
#include "Utility.h"

msl::test::fft::complex msl::test::fft::operator+(msl::test::fft::complex c1, msl::test::fft::complex c2) {
	return msl::test::fft::complex(c1.real + c2.real, c1.imag + c2.imag);
}

msl::test::fft::complex msl::test::fft::operator*(msl::test::fft::complex c1, msl::test::fft::complex c2) {
	return msl::test::fft::complex(c1.real * c2.real - c1.imag * c2.imag, c1.real * c2.imag +
		c1.imag * c2.real);
}

std::ostream& msl::test::fft::operator<<(std::ostream& os, const msl::test::fft::complex& c) {
	os << c.real << " + i*" << c.imag;

	return os;
}

msl::test::fft::complex msl::test::fft::copy(const msl::DistributedArray<complex>& A, int i,
											   msl::test::fft::complex Ai) {
	return A.get(i);
}

msl::test::fft::complex msl::test::fft::init(const msl::DistributedArray<double>& A, int i) {
	return msl::test::fft::complex(A.get(i), 0.0);
}

int msl::test::fft::bitcomplement(int k, int i) {
	return i ^ (1 << k);
}

msl::test::fft::complex msl::test::fft::omega(int j, int i) {
	int b = i >> (log2size - j - 1);
	int b2 = 0;

	for(int l = 0;l <= j;l++) {
		b2 = (b & 1) ? 2 * b2 + 1 : 2 * b2;
		b >>= 1;
	}

	double v = 2.0 * pi / n * (b2 << (log2size - j - 1));
	
	return msl::test::fft::complex(cos(v), sin(v));
}

msl::test::fft::complex msl::test::fft::combine(const msl::DistributedArray<msl::test::fft::complex>& T,
												  int j, int i, msl::test::fft::complex Ai) {
	return (i & (1 << log2size - 1 - j)) ? T.get(i) + omega(j, i) * Ai : Ai + omega(j, i) * T.get(i);
}

msl::test::fft::complex msl::test::fft::fetch(const msl::DistributedArray<msl::test::fft::complex>& R,
												int j, int i, msl::test::fft::complex Ti) {
	return R.get(bitcomplement(log2size - 1 - j, i));
}

msl::DistributedArray<msl::test::fft::complex> msl::test::fft::fft(const msl::DistributedArray<double>& A, int n) {
	int j;
	msl::DistributedArray<msl::test::fft::complex> R(n, curry(init)(A));
	msl::DistributedArray<msl::test::fft::complex> T(n);

	for(j = 0; j < log2p; j++) {
		T.mapIndexInPlace(curry(copy)(R));
		T.permutePartition(curry(bitcomplement)(log2p - 1 - j));
		R.mapIndexInPlace(curry(combine)(T)(j));
	}

	for(j = log2p; j < log2size; j++) {
		T.mapIndexInPlace(curry(fetch)(R)(j));
		R.mapIndexInPlace(curry(combine)(T)(j));
	}
	return R;
}

// *********************************************************************************
// Test: FFT (algorithm (roughly) according to:
// M. Quinn: Parallel Computing - Theory and Practice, McGraw-Hill, 1994, p. 207)
//
// - n determines the size of the distributed array to transform; note that size
//   must be a power of 2
// - nt determines the number of processor cores to use for the transformation
// *********************************************************************************
void msl::test::fft::testFft(int n, int nt, bool flag) {
	msl::utl::printv("Welcome to the test suite for the fast fourier transformation.\n");

	// 2^30: 1073741824
	// 2^28: 268435456; np = 1: ~20 minutes
	// 2^26: 67108864;  np = 2: ~4,5 minutes
	// 2^24: 16777216
	// 2^20: 1048576

	np = Muesli::MSL_numOfTotalProcs;
	log2p = log2(np);
	log2size = log2(n);
	msl::oal::setNumThreads(nt);

	msl::DistributedArray<double> A(n, 1.0);

	if(flag) {
		msl::utl::printv("A = ");
		A.show();
		msl::utl::printv("\n");
	}

	msl::DistributedArray<complex> B = fft(A, n);

	if(flag) {
		msl::utl::printv("B = ");
		B.show();
	}

	msl::utl::printv("\nAll tests successfully passed!\n\n");
}


int main(int argc, char** argv) {
	msl::InitSkeletons(argc, argv);
	msl::test::fft::testFft(16, 2, true);
	msl::TerminateSkeletons();
	return 0;
}
