// **********************************************************************************************************
// *																										*
// * This Software is published under the MIT License.														*
// * For more information and the license text, please refer to the file 'License' located in the zip file. *
// *																										*
// * Copyright (c) 2001-2011 Steffen Ernsting, Philipp Ciechanowicz, Michael Poldner and Herbert Kuchen     *                 
// *        		{s.ernsting|ciechanowicz|poldner|kuchen}@uni-muenster.de              					*
// *																										*	
// **********************************************************************************************************

#include "mpi.h"

#include "DistributedSparseMatrix.h"
#include "MatrixMarketDistributedSparseMatrixReader.h"
#include "MatrixMarketMatrixReader.h"
#include "Muesli.h"
#include "OAL.h"
#include "TestMisc.h"
#include "Utility.h"

// AUXILIARY

// AUXILIARY / ARGUMENT FUNCTIONS

double msl::test::misc::add(double a, double b) {
	return a + b;
}

double msl::test::misc::addIndex(double a, double b, int i, int j) {
	return a + b;
}

// TEST FUNCTIONS

/* Compares the runtime of the zipIndexInPlace skeleton. One version
   uses critical pragmas to synchronize the access to the non-const
   member function addSubmatrix, the other uses an array of vectors
   to store the newly created submatrices.

   The goal is to construct a scenario, where the non-synchronized
   version of the zipIndexInPlace skeleton is faster than the one
   which uses critical directives. The test case is constructed as
   follows: Matrix m1 is completely empty, whereas matrix m2 is
   completey filled. Thus, the maximum number of calls to addSubmatrix
   should be made. In addition, the number of submatrices is very
   high (100,000 * 100,000 / (10 * 10)) = 100,000,000.

   Results: zipIndexInPlaceNoCritical indeed performs better in this
   test scenario.
*/
void msl::test::misc::testZipIndexInPlace() {
	utl::printv("Testing DSM zipIndexInPlace vs. zipIndexInPlace critical...\n");

	n = 100;
	m = 100;
	r = 10;
	c = 10;
	runs = 10;
	msl::DistributedSparseMatrix<> m1(n, m, r, c, 0);
	msl::DistributedSparseMatrix<> m2(n, m, r, c, 0);

	for(int i = 0; i < n; i++) {
		m2.setElement(1, n - i - 1, i);
	}
	
	nt = Muesli::MSL_ARG1;
	msl::oal::setNumThreads(nt);
		
	t0 = MPI_Wtime();
	
	for(int i = 0; i < runs; i++) {
		msl::DistributedSparseMatrix<> m3(m1);
		msl::DistributedSparseMatrix<> m4(m2);
		m3.zipIndexInPlace(m4, &addIndex);
	}

	t1 = MPI_Wtime();
	
	for(int i = 0; i < runs; i++) {
		msl::DistributedSparseMatrix<> m3(m1);
		msl::DistributedSparseMatrix<> m4(m2);
		m3.zipIndexInPlaceCritical(m4, &addIndex);
	}

	t2 = MPI_Wtime();
	
	for(int i = 0; i < runs; i++) {
		msl::DistributedSparseMatrix<> m3(m1);
		msl::DistributedSparseMatrix<> m4(m2);
		m3.zipIndexInPlace(m4, &addIndex);
	}

	t3 = MPI_Wtime();
	
	for(int i = 0; i < runs; i++) {
		msl::DistributedSparseMatrix<> m3(m1);
		msl::DistributedSparseMatrix<> m4(m2);
		m3.zipIndexInPlaceCritical(m4, &addIndex);
	}

	t4 = MPI_Wtime();
	
	for(int i = 0; i < runs; i++) {
		msl::DistributedSparseMatrix<> m3(m1);
		msl::DistributedSparseMatrix<> m4(m2);
		m4.zipIndexInPlaceCritical(m3, &addIndex);
	}

	t5 = MPI_Wtime();
	
	for(int i = 0; i < runs; i++) {
		msl::DistributedSparseMatrix<> m3(m1);
		msl::DistributedSparseMatrix<> m4(m2);
		m4.zipIndexInPlace(m3, &addIndex);
	}

	t6 = MPI_Wtime();

	utl::printv("nt = %d\nm1.ziip(m2):  %f\nm1.ziipc(m2): %f\nm1.ziip(m2):  %f\nm1.ziipc(m2): %f\n" \
		"m2.ziipc(m1): %f\nm2.ziip(m1):  %f\n\n", nt, t1 - t0, t2 - t1, t3 - t2, t4 - t3, t5 - t4, t6 - t5);
}

/* Compares the runtime of multiply and multiplyCritical.
*/
void msl::test::misc::testMultiply() {
	msl::utl::printv("Testing DSM multiply vs. multiply critical...\n");

	msl::MatrixElement<double> val;
	n  = 500;
	m  = 500;
	r  = 50;
	c  = 50;
	np = Muesli::MSL_numOfTotalProcs;
	nt = Muesli::MSL_ARG1;
	double* b = new double[n];
	double* x = new double[m];
	msl::DistributedSparseMatrix<double> dsm(n, m, r, c, 0);

	for(int i = 0; i < n; i++) {
		b[i] = 0;
	}

	for(int i = 0; i < m; i++) {
		x[i] = 1;
	}

	for(int i = 0; i < n; i++) {
		for(int j = 0; j < m; j++) {
			dsm.setElement(1, i, j);
		}
	}

	msl::oal::setNumThreads(nt);

	t0 = MPI_Wtime();

	dsm.multiply(x, b);

	t1 = MPI_Wtime();

	dsm.multiplyCritical(x, b);

	t2 = MPI_Wtime();

	msl::utl::printv("np = %d, nt = %d\nmlt    = %f\nmltcrt = %f\n\n",
		np, nt, t1 - t0, t2 - t1);
}

/* Compares the runtime of different copy constructors.
*/
void msl::test::misc::testConstructors() {
	n = 500;
	m = 500;
	r = 10;
	c = 10;
	runs = 10;

	nt = Muesli::MSL_ARG1;

	msl::utl::printv("Testing DSM constructors...\n");

	msl::DistributedSparseMatrix<int> m1(n, m, r, c, 0);

	for(int i = 0; i < n; i++) {
		for(int j = 0; j < m; j++) {
			m1.setElement(1, i, j);
		}
	}

	msl::oal::setNumThreads(nt);

	t0 = MPI_Wtime();

	for(int i = 0; i < runs; i++) {
		msl::DistributedSparseMatrix<int> m2(m1);
	}

	t1 = MPI_Wtime();

	for(int i = 0; i < runs; i++) {
		msl::DistributedSparseMatrix<int> m2(m1);
	}

	t2 = MPI_Wtime();

	for(int i = 0; i < runs; i++) {
		msl::DistributedSparseMatrix<int> m2(m1, 0);
	}

	t3 = MPI_Wtime();

	msl::oal::setNumThreads(1);

	for(int i = 0; i < runs; i++) {
		msl::DistributedSparseMatrix<int> m2(m1);
	}

	t4 = MPI_Wtime();

	for(int i = 0; i < runs; i++) {
		msl::DistributedSparseMatrix<int> m2(m1, 0);
	}

	t5 = MPI_Wtime();

	msl::utl::printv("m2(m1):    %f\nm2(m1):    %f\nm2(m1, 0): %f\nm2(m1):    %f\nm2(m1, 0): %f\n\n",
		t1 - t0, t2 - t1, t3 - t2, t4 - t3, t5 - t4);
}

/* Function must only be executed with at least np = 4 processes.
*/
void msl::test::misc::testGroups() {
	msl::utl::printv("Testing MPI group management... ");

	MPI_Comm c3;
	MPI_Group g0, g1, g2, g3;
	int a[] = {0, 1};
	int b[] = {0, 3};

	MPI_Comm_group(MPI_COMM_WORLD, &g0);
	MPI_Group_incl(g0, 2, a, &g1);
	MPI_Group_excl(g0, 2, b, &g2);
	MPI_Group_union(g1, g2, &g3);
	MPI_Comm_create(MPI_COMM_WORLD, g3, &c3);

	msl::utl::printv("done!\n\n");
}

/* Function must only be executed with at least np = 4 processes.
*/
void msl::test::misc::testMpiAllgather() {
	id = Muesli::MSL_myId;
	np = Muesli::MSL_numOfTotalProcs;
	int  sbuf = id + 3;
	int* rbuf = new int[np];

	for(int i = 0; i < np; i++) {
		rbuf[i] = 0;
	}

	msl::utl::printv("Testing MPI_Allgather:\n");

	MPI_Barrier(MPI_COMM_WORLD);

	printf("id = %d, sbuf = %d\n", id, sbuf);

	msl::utl::print(rbuf, np);

	MPI_Allgather(&sbuf, 1, MPI_INT, rbuf, 1, MPI_INT, MPI_COMM_WORLD);

	printf("id = %d, sbuf = %d\n", id, sbuf);

	MPI_Barrier(MPI_COMM_WORLD);

	msl::utl::print(rbuf, np);

	msl::utl::printv("\n");
}

/* Function must only be executed with at least np = 3 processes.
*/
void msl::test::misc::testMpiAllreduce() {
	id = Muesli::MSL_myId;
	np = Muesli::MSL_numOfTotalProcs;
	int sbuf = id + 3;
	int rbuf = 0;

	msl::utl::printv("Testing MPI_Allreduce:\n");

	MPI_Barrier(MPI_COMM_WORLD);

	printf("id = %d, sbuf = %d, rbuf = %d\n", id, sbuf, rbuf);

	MPI_Barrier(MPI_COMM_WORLD);

	MPI_Allreduce(&sbuf, &rbuf, 1, MPI_INT, MPI_SUM, MPI_COMM_WORLD);

	printf("id = %d, sbuf = %d, rbuf = %d\n", id, sbuf, rbuf);

	MPI_Barrier(MPI_COMM_WORLD);

	msl::utl::printv("\n");
}

/* Function must only be executed with at least np = 3 processes.
*/
void msl::test::misc::testMpiBroadcast() {
	id = Muesli::MSL_myId;
	int buf = id + 3;

	msl::utl::printv("Testing MPI_Bcast:\n");

	MPI_Barrier(MPI_COMM_WORLD);

	printf("id = %d, buf = %d\n", id, buf);

	MPI_Barrier(MPI_COMM_WORLD);

	MPI_Bcast(&buf, 1, MPI_INT, 2, MPI_COMM_WORLD);

	printf("id = %d, buf = %d\n", id, buf);

	MPI_Barrier(MPI_COMM_WORLD);

	msl::utl::printv("\n");
}

/* Function must only be executed with at least np = 3 processes.
*/
void msl::test::misc::testMpiGather() {
	id = Muesli::MSL_myId;
	np = Muesli::MSL_numOfTotalProcs;
	int  sbuf = id + 3;
	int* rbuf = new int[np];

	for(int i = 0; i < np; i++) {
		rbuf[i] = 0;
	}

	msl::utl::printv("Testing MPI_Gather:\n");

	MPI_Barrier(MPI_COMM_WORLD);

	printf("id = %d, sbuf = %d\n", id, sbuf);

	MPI_Barrier(MPI_COMM_WORLD);

	msl::utl::print(rbuf, np);

	MPI_Gather(&sbuf, 1, MPI_INT, rbuf, 1, MPI_INT, 2, MPI_COMM_WORLD);

	printf("id = %d, sbuf = %d\n", id, sbuf);

	MPI_Barrier(MPI_COMM_WORLD);

	msl::utl::print(rbuf, np);

	msl::utl::printv("\n");
}

/* Function must only be executed with at least np = 3 processes.
*/
void msl::test::misc::testMpiReduce() {
	id = Muesli::MSL_myId;
	np = Muesli::MSL_numOfTotalProcs;
	int sbuf = id + 3;
	int rbuf = 0;

	msl::utl::printv("Testing MPI_Reduce:\n");

	MPI_Barrier(MPI_COMM_WORLD);

	printf("id = %d, sbuf = %d, rbuf = %d\n", id, sbuf, rbuf);

	MPI_Barrier(MPI_COMM_WORLD);

	MPI_Reduce(&sbuf, &rbuf, 1, MPI_INT, MPI_SUM, 2, MPI_COMM_WORLD);

	printf("id = %d, sbuf = %d, rbuf = %d\n", id, sbuf, rbuf);

	MPI_Barrier(MPI_COMM_WORLD);

	msl::utl::printv("\n");
}

// TEST FUNCTIONS / VARIOUS

void msl::test::misc::testWriteRead() {
	utl::printv("Testing read/write:\n");

	bool   b = true;
	char   c = 'c';
	short  s = 2;
	int    i = 4;
	long   l = 8;
	float  f = 16.32;
	double d = 64.128;

	//char* buffer = new char[msl::SOB + msl::SOC + msl::SOS + msl::SOI + msl::SOL + msl::SOF + msl::SOD];
	void* buffer = new char[msl::SOB + msl::SOC + msl::SOS + msl::SOI + msl::SOL + msl::SOF + msl::SOD];

	msl::write(buffer, b, 0);
	msl::write(buffer, c, msl::SOB);
	msl::write(buffer, s, msl::SOB + msl::SOC);
	msl::write(buffer, i, msl::SOB + msl::SOC + msl::SOS);
	msl::write(buffer, l, msl::SOB + msl::SOC + msl::SOS + msl::SOI);
	msl::write(buffer, f, msl::SOB + msl::SOC + msl::SOS + msl::SOI + msl::SOL);
	msl::write(buffer, d, msl::SOB + msl::SOC + msl::SOS + msl::SOI + msl::SOL + msl::SOF);

	utl::printv("buffer[%d] = %d \n", 0,
		msl::read<bool>(buffer, 0));
	utl::printv("buffer[%d] = %c \n", msl::SOB,
		msl::read<char>(buffer, msl::SOB));
	utl::printv("buffer[%d] = %hu\n", msl::SOB + msl::SOC,
		msl::read<short>(buffer, msl::SOB + msl::SOC));
	utl::printv("buffer[%d] = %d \n", msl::SOB + msl::SOC + msl::SOS,
		msl::read<int>(buffer, msl::SOB + msl::SOC + msl::SOS));
	utl::printv("buffer[%d] = %u \n", msl::SOB + msl::SOC + msl::SOS + msl::SOI,
		msl::read<long>(buffer, msl::SOB + msl::SOC + msl::SOS + msl::SOI));
	utl::printv("buffer[%d] = %f \n", msl::SOB + msl::SOC + msl::SOS + msl::SOI + msl::SOL,
		msl::read<float>(buffer, msl::SOB + msl::SOC + msl::SOS + msl::SOI + msl::SOL));
	utl::printv("buffer[%d] = %f \n\n", msl::SOB + msl::SOC + msl::SOS + msl::SOI + msl::SOL + msl::SOF,
		msl::read<double>(buffer, msl::SOB + msl::SOC + msl::SOS + msl::SOI + msl::SOL + msl::SOF));
}



int main(int argc, char** argv) {
	msl::InitSkeletons(argc, argv);
	msl::test::misc::testConstructors();
	msl::test::misc::testGroups();
	msl::test::misc::testMpiAllgather();
	msl::test::misc::testMpiBroadcast();
	msl::test::misc::testMpiGather();
	msl::test::misc::testMpiReduce();
	msl::test::misc::testMultiply();
	msl::test::misc::testWriteRead();
	msl::test::misc::testZipIndexInPlace();
	msl::TerminateSkeletons();
	return 0;
}

