// **********************************************************************************************************
// *																										*
// * This Software is published under the MIT License.														*
// * For more information and the license text, please refer to the file 'License' located in the zip file. *
// *																										*
// * Copyright (c) 2001-2011 Steffen Ernsting, Philipp Ciechanowicz, Michael Poldner and Herbert Kuchen     *                 
// *        		{s.ernsting|ciechanowicz|poldner|kuchen}@uni-muenster.de              					*
// *																										*	
// **********************************************************************************************************

#include <iostream>
#include <math.h>

#include "DistributedMatrix.h"
#include "Muesli.h"
#include "OAL.h"
#include "TestShortest.h"
#include "Utility.h"

inline int msl::test::shortest::negate(const int a) {
	return -a;
}

template <class C>
inline int msl::test::shortest::copy(const msl::DistributedMatrix<C>& A, int row, int col, C x) {
	return A.getLocal(row,col, true);
}

inline int msl::test::shortest::min(const int a, const int b) {
	return a<b ? a : b;
}

const int infinity = 999999;

inline int msl::test::shortest::init(int x, int y) {
	return abs(x - y) <= 1 ? abs(x - y) : infinity;
}

template<class C> 
C msl::test::shortest::skprod(const msl::DistributedMatrix<C>& A,
												const msl::DistributedMatrix<C>& B,
												int i, int j, C Cij) {
	C sum = Cij;

	for(int k = 0; k < A.getLocalRows(); ++k)
		//sum += A.getGlobalLocal(i,k) * B.getLocalGlobal(k,j);
		sum = min(sum, B.getLocalGlobal(k, j) + A.getGlobalLocal(i, k));

	return sum;
}

/* Using algorithm of Gentleman based on torus topology
*/
template<class C> 
msl::DistributedMatrix<C> msl::test::shortest::matmult(msl::DistributedMatrix<C> A, msl::DistributedMatrix<C> B) {
	// assumption: A, B have same square shape
	A.rotateRows(&negate);
	B.rotateCols(&negate);

	msl::DistributedMatrix<C> R(A.getRows(), A.getCols(), infinity, A.getBlocksInCol(), A.getBlocksInRow());

	for(int i = 0; i < A.getBlocksInRow(); ++i) {
		typedef C (*skprod_t)(const msl::DistributedMatrix<C>&,
			const msl::DistributedMatrix<C>&, int, int, C);
		R.mapIndexInPlace(curry((skprod_t)skprod<C>)(A)(B));
        A.rotateRows(-1);
		B.rotateCols(-1);
	}

	return R;
}


template <class C> // assumption: Problemsize is power of 2
msl::DistributedMatrix<C> msl::test::shortest::shortestPaths(msl::DistributedMatrix<C>& A, int n) {
	msl::DistributedMatrix<C> B(A.getRows(), A.getCols(), 0, A.getBlocksInCol(), A.getBlocksInRow());
	
	for(int i = 0; i < log2(n); i++) {
		typedef int (*copy_t)(const msl::DistributedMatrix<C>&, int, int, C);
		B.mapIndexInPlace(curry((copy_t)copy<C>)(A)); // B is copy of A 
		A = matmult(A, B);
	}
	return A;
}

// *************************************************************
// Test: matrix multiplication
// *************************************************************

void msl::test::shortest::testShortest(int n, bool flag) {
	// square root of np
	int sqrtp;
	// calculate square root of np
	sqrtp = (int)(sqrt((double)Muesli::MSL_numOfTotalProcs) + 0.1);
	// create matrix A
	msl::DistributedMatrix<int> A(n, n, &init, sqrtp, sqrtp);
	
	// output matrix A
	if(flag) {
		msl::utl::printv("A =\n");
		A.show();
		msl::utl::printv("\n");
	}

	msl::DistributedMatrix<int> B =	shortestPaths(A, n);

	// output matrix C
	if(flag) {
        msl::utl::printv("ShortestPaths(A) =\n");
		B.show();
	}
}


int main(int argc, char** argv) {
	msl::InitSkeletons(argc, argv);
	if (msl::Muesli::MSL_ARG1 > 0)
		msl::test::shortest::testShortest(msl::Muesli::MSL_ARG1, false);
	else
		msl::test::shortest::testShortest(8, true);
	msl::TerminateSkeletons();
	return 0;
}
